@@ -31,42 +31,65 @@ struct extract_args {
 	int ret;
 };
 
-static struct snapshot_hdr snaphdr;
+static struct snapshot_hdr snap_hdr;
+static struct blk_hdr blk_hdr;
 static struct cache *cache;
 static int ifd;
 static int sfd;
 static int cfd;
-static int verbose;
 static int cache_dirty;
 static unsigned long long cache_hits;
 static unsigned long long cache_misses;
+
+int verbose;
 char *argv0;
 
 static size_t
-comp_size(size_t size)
+compr_size(size_t size)
 {
-	return LZ4_compressBound(size);
-}
+	size_t ret;
 
-static size_t
-comp(uint8_t *in, uint8_t *out, size_t insize, size_t outsize)
-{
-	int ret;
+	if (compr_enabled)
+		ret = LZ4_compressBound(size);
+	else
+		ret = size;
 
-	ret = LZ4_compress_default((char *)in, (char *)out, insize, outsize);
-	if (ret < 0)
-		errx(1, "LZ4_compress_default failed");
 	return ret;
 }
 
 static size_t
-decomp(uint8_t *in, uint8_t *out, size_t insize, size_t outsize)
+compr(uint8_t *in, uint8_t *out, size_t insize, size_t outsize)
 {
 	int ret;
 
-	ret = LZ4_decompress_safe((char *)in, (char *)out, insize, outsize);
-	if (ret < 0)
-		errx(1, "LZ4_decompress_safe failed");
+	if (compr_enabled) {
+		ret = LZ4_compress_default((char *)in, (char *)out, insize,
+		                           outsize);
+		if (ret < 0)
+			errx(1, "LZ4_compress_default failed");
+	} else {
+		ret = insize;
+		memcpy(out, in, insize);
+	}
+
+	return ret;
+}
+
+static size_t
+decompr(uint8_t *in, uint8_t *out, size_t insize, size_t outsize)
+{
+	int ret;
+
+	if (compr_enabled) {
+		ret = LZ4_decompress_safe((char *)in, (char *)out, insize,
+		                          outsize);
+		if (ret < 0)
+			errx(1, "LZ4_decompress_safe failed");
+	} else {
+		ret = insize;
+		memcpy(out, in, insize);
+	}
+
 	return ret;
 }
 
@@ -88,7 +111,7 @@ print_stats(struct stats *st)
 	fprintf(stderr, "Original size: %llu bytes\n",
 	        (unsigned long long)st->orig_size);
 	fprintf(stderr, "Compressed size: %llu bytes\n",
-	        (unsigned long long)st->comp_size);
+	        (unsigned long long)st->compr_size);
 	fprintf(stderr, "Deduplicated size: %llu bytes\n",
 	        (unsigned long long)st->dedup_size);
 	fprintf(stderr, "Min/avg/max block size: %llu/%llu/%llu bytes\n",
@@ -104,16 +127,22 @@ print_stats(struct stats *st)
 static void
 append_snap(struct snapshot *snap)
 {
-	/* Update snapshot header */
-	snaphdr.nr_snapshots++;
-	xlseek(ifd, 0, SEEK_SET);
-	xwrite(ifd, &snaphdr, sizeof(snaphdr));
+	if (mul_overflow(snap->nr_blk_descs, BLK_DESC_SIZE))
+		errx(1, "%s: overflow", __func__);
+	snap->size = snap->nr_blk_descs * BLK_DESC_SIZE;
 
-	/* Append snapshot */
-	xlseek(ifd, 0, SEEK_END);
-	snap->size = sizeof(*snap);
-	snap->size += snap->nr_blk_descs * sizeof(snap->blk_desc[0]);
-	xwrite(ifd, snap, snap->size);
+	if (add_overflow(SNAPSHOT_SIZE, snap->size))
+		errx(1, "%s: overflow", __func__);
+	snap->size += SNAPSHOT_SIZE;
+
+	xlseek(ifd, snap_hdr.size, SEEK_SET);
+	write_snapshot(ifd, snap);
+	write_snapshot_blk_descs(ifd, snap);
+
+	if (add_overflow(snap_hdr.size, snap->size))
+		errx(1, "%s: overflow", __func__);
+	snap_hdr.size += snap->size;
+	snap_hdr.nr_snapshots++;
 }
 
 static struct snapshot *
@@ -138,8 +167,14 @@ grow_snap(struct snapshot *snap, uint64_t nr_blk_descs)
 {
 	size_t size;
 
-	size = sizeof(*snap);
-	size += nr_blk_descs * sizeof(snap->blk_desc[0]);
+	if (mul_overflow(nr_blk_descs, sizeof(snap->blk_desc[0])))
+		errx(1, "%s: overflow", __func__);
+	size = nr_blk_descs * sizeof(snap->blk_desc[0]);
+
+	if (add_overflow(size, sizeof(*snap)))
+		errx(1, "%s: overflow", __func__);
+	size += sizeof(*snap);
+
 	snap = realloc(snap, size);
 	if (snap == NULL)
 		err(1, "realloc");
@@ -184,9 +219,9 @@ read_blk(uint8_t *buf, struct blk_desc *blk_desc)
 static void
 append_blk(uint8_t *buf, struct blk_desc *blk_desc)
 {
-	xlseek(sfd, snaphdr.store_size, SEEK_SET);
+	xlseek(sfd, blk_hdr.size, SEEK_SET);
 	xwrite(sfd, buf, blk_desc->size);
-	snaphdr.store_size += blk_desc->size;
+	blk_hdr.size += blk_desc->size;
 }
 
 static void
@@ -194,27 +229,27 @@ dedup_chunk(struct snapshot *snap, uint8_t *chunkp, size_t chunk_size)
 {
 	uint8_t md[MDSIZE];
 	struct cache_entry cache_entry;
-	uint8_t *comp_buf;
+	uint8_t *compr_buf;
 	size_t n;
 
-	comp_buf = alloc_buf(comp_size(BLKSIZE_MAX));
+	compr_buf = alloc_buf(compr_size(BLKSIZE_MAX));
 
-	n = comp(chunkp, comp_buf, chunk_size, comp_size(BLKSIZE_MAX));
-	hash_blk(comp_buf, n, md);
+	n = compr(chunkp, compr_buf, chunk_size, compr_size(BLKSIZE_MAX));
+	hash_blk(compr_buf, n, md);
 
-	snaphdr.st.orig_size += chunk_size;
-	snaphdr.st.comp_size += n;
+	snap_hdr.st.orig_size += chunk_size;
+	snap_hdr.st.compr_size += n;
 
 	memcpy(cache_entry.md, md, sizeof(cache_entry.md));
 	if (lookup_cache_entry(cache, &cache_entry) < 0) {
 		struct blk_desc blk_desc;
 
 		memcpy(&blk_desc.md, md, sizeof(blk_desc.md));
-		blk_desc.offset = snaphdr.store_size;
+		blk_desc.offset = blk_hdr.size;
 		blk_desc.size = n;
 
 		snap->blk_desc[snap->nr_blk_descs++] = blk_desc;
-		append_blk(comp_buf, &blk_desc);
+		append_blk(compr_buf, &blk_desc);
 
 		cache_entry.offset = blk_desc.offset;
 		cache_entry.size = blk_desc.size;
@@ -222,13 +257,13 @@ dedup_chunk(struct snapshot *snap, uint8_t *chunkp, size_t chunk_size)
 		add_cache_entry(cache, &cache_entry);
 		cache_misses++;
 
-		snaphdr.st.dedup_size += blk_desc.size;
-		snaphdr.st.nr_blks++;
+		snap_hdr.st.dedup_size += blk_desc.size;
+		snap_hdr.st.nr_blks++;
 
-		if (blk_desc.size > snaphdr.st.max_blk_size)
-			snaphdr.st.max_blk_size = blk_desc.size;
-		if (blk_desc.size < snaphdr.st.min_blk_size)
-			snaphdr.st.min_blk_size = blk_desc.size;
+		if (blk_desc.size > snap_hdr.st.max_blk_size)
+			snap_hdr.st.max_blk_size = blk_desc.size;
+		if (blk_desc.size < snap_hdr.st.min_blk_size)
+			snap_hdr.st.min_blk_size = blk_desc.size;
 	} else {
 		struct blk_desc blk_desc;
 
@@ -239,7 +274,7 @@ dedup_chunk(struct snapshot *snap, uint8_t *chunkp, size_t chunk_size)
 		cache_hits++;
 	}
 
-	free(comp_buf);
+	free(compr_buf);
 }
 
 static void
@@ -251,7 +286,7 @@ dedup(int fd, char *msg)
 	ssize_t n;
 
 	snap = alloc_snap();
-	chunker = alloc_chunker(BLKSIZE_MAX, fd);
+	chunker = alloc_chunker(fd, BLKSIZE_MAX);
 
 	SHA256_Init(&ctx);
 	while ((n = fill_chunker(chunker)) > 0) {
@@ -295,14 +330,14 @@ extract(struct snapshot *snap, void *arg)
 		return WALK_CONTINUE;
 
 	buf[0] = alloc_buf(BLKSIZE_MAX);
-	buf[1] = alloc_buf(comp_size(BLKSIZE_MAX));
+	buf[1] = alloc_buf(compr_size(BLKSIZE_MAX));
 	for (i = 0; i < snap->nr_blk_descs; i++) {
 		struct blk_desc *blk_desc;
 		size_t blksize;
 
 		blk_desc = &snap->blk_desc[i];
 		read_blk(buf[1], blk_desc);
-		blksize = decomp(buf[1], buf[0], blk_desc->size, BLKSIZE_MAX);
+		blksize = decompr(buf[1], buf[0], blk_desc->size, BLKSIZE_MAX);
 		xwrite(args->fd, buf[0], blksize);
 	}
 	free_buf(buf[1]);
@@ -319,7 +354,7 @@ check(struct snapshot *snap, void *arg)
 	SHA256_CTX ctx;
 	uint64_t i;
 
-	buf = alloc_buf(comp_size(BLKSIZE_MAX));
+	buf = alloc_buf(compr_size(BLKSIZE_MAX));
 	/*
 	 * Calculate hash for each block and compare
 	 * against snapshot entry block descriptor
@@ -372,7 +407,7 @@ rebuild_cache(struct snapshot *snap, void *arg)
 	SHA256_CTX ctx;
 	uint64_t i;
 
-	buf = alloc_buf(comp_size(BLKSIZE_MAX));
+	buf = alloc_buf(compr_size(BLKSIZE_MAX));
 	for (i = 0; i < snap->nr_blk_descs; i++) {
 		struct cache_entry cache_entry;
 		struct blk_desc *blk_desc;
@@ -387,10 +422,10 @@ rebuild_cache(struct snapshot *snap, void *arg)
 		memcpy(cache_entry.md, blk_desc->md, sizeof(cache_entry.md));
 		cache_entry.offset = blk_desc->offset;
 		cache_entry.size = blk_desc->size;
-		cache_dirty = 1;
 		add_cache_entry(cache, &cache_entry);
 	}
 	free(buf);
+	cache_dirty = 1;
 	return WALK_CONTINUE;
 }
 
@@ -400,19 +435,14 @@ walk_snap(int (*fn)(struct snapshot *, void *), void *arg)
 {
 	uint64_t i;
 
-	xlseek(ifd, sizeof(snaphdr), SEEK_SET);
-	for (i = 0; i < snaphdr.nr_snapshots; i++) {
+	for (i = 0; i < snap_hdr.nr_snapshots; i++) {
 		struct snapshot *snap;
 		int ret;
 
 		snap = alloc_snap();
-		if (xread(ifd, snap, sizeof(*snap)) == 0)
-			errx(1, "read: unexpected EOF");
-
+		read_snapshot(ifd, snap);
 		snap = grow_snap(snap, snap->nr_blk_descs);
-		if (xread(ifd, snap->blk_desc,
-		          snap->nr_blk_descs * sizeof(snap->blk_desc[0])) == 0)
-			errx(1, "read: unexpected EOF");
+		read_snapshot_descs(ifd, snap);
 
 		ret = (*fn)(snap, arg);
 		free(snap);
@@ -421,53 +451,135 @@ walk_snap(int (*fn)(struct snapshot *, void *), void *arg)
 	}
 }
 
+static void
+match_ver(uint64_t v)
+{
+	uint8_t maj, min;
+
+	min = v & VER_MIN_MASK;
+	maj = (v >> VER_MAJ_SHIFT) & VER_MAJ_MASK;
+	if (maj == VER_MAJ && min == VER_MIN)
+		return;
+	errx(1, "format version mismatch: expected %u.%u but got %u.%u",
+	     VER_MAJ, VER_MIN, maj, min);
+}
+
 static int
 flush_cache(struct cache_entry *cache_entry)
 {
-	xwrite(cfd, cache_entry, sizeof(*cache_entry));
+	write_cache_entry(cfd, cache_entry);
 	return 0;
 }
 
-static uint64_t
-cache_nr_entries(void)
-{
-	struct stat sb;
-
-	if (fstat(cfd, &sb) < 0)
-		err(1, "fstat");
-	return sb.st_size / sizeof(struct cache_entry);
-}
-
 static void
 load_cache(void)
 {
+	struct stat sb;
 	uint64_t nr_entries;
 	uint64_t i;
 
 	xlseek(cfd, 0, SEEK_SET);
-	nr_entries = cache_nr_entries();
+
+	if (fstat(cfd, &sb) < 0)
+		err(1, "fstat");
+
+	nr_entries = sb.st_size / CACHE_ENTRY_SIZE;
+	if (nr_entries == 0) {
+		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
+		walk_snap(rebuild_cache, NULL);
+		return;
+	}
+
 	for (i = 0; i < nr_entries; i++) {
 		struct cache_entry cache_entry;
 
-		if (xread(cfd, &cache_entry, sizeof(cache_entry)) == 0)
-			errx(1, "read: unexpected EOF");
+		read_cache_entry(cfd, &cache_entry);
 		add_cache_entry(cache, &cache_entry);
 	}
 }
 
 static void
-init(void)
+save_cache(void)
 {
-	struct stat sb;
+	if (cache_dirty) {
+		xlseek(cfd, 0, SEEK_SET);
+		walk_cache(cache, flush_cache);
+	}
+}
 
-	ifd = open(SNAPSF, O_RDWR | O_CREAT, 0600);
+static void
+init_blk_hdr(void)
+{
+	blk_hdr.flags = (VER_MAJ << VER_MAJ_SHIFT) | VER_MIN;
+	blk_hdr.flags |= compr_enabled << COMPR_ENABLED_SHIFT;
+	blk_hdr.size = BLK_HDR_SIZE;
+}
+
+static void
+load_blk_hdr(void)
+{
+	uint64_t v;
+
+	xlseek(sfd, 0, SEEK_SET);
+	read_blk_hdr(sfd, &blk_hdr);
+	match_ver(blk_hdr.flags);
+
+	v = blk_hdr.flags >> COMPR_ENABLED_SHIFT;
+	v &= COMPR_ENABLED_MASK;
+	compr_enabled = v;
+}
+
+static void
+save_blk_hdr(void)
+{
+	xlseek(sfd, 0, SEEK_SET);
+	write_blk_hdr(sfd, &blk_hdr);
+}
+
+static void
+init_snap_hdr(void)
+{
+	snap_hdr.flags = (VER_MAJ << VER_MAJ_SHIFT) | VER_MIN;
+	snap_hdr.size = SNAP_HDR_SIZE;
+	snap_hdr.st.min_blk_size = compr_size(BLKSIZE_MAX);
+}
+
+static void
+load_snap_hdr(void)
+{
+	xlseek(ifd, 0, SEEK_SET);
+	read_snap_hdr(ifd, &snap_hdr);
+	match_ver(snap_hdr.flags);
+}
+
+static void
+save_snap_hdr(void)
+{
+	xlseek(ifd, 0, SEEK_SET);
+	write_snap_hdr(ifd, &snap_hdr);
+}
+
+static void
+init(int iflag)
+{
+	int flags;
+
+	flags = O_RDWR;
+	if (iflag)
+		flags |= O_CREAT | O_EXCL;
+
+	ifd = open(SNAPSF, flags, 0600);
 	if (ifd < 0)
 		err(1, "open %s", SNAPSF);
 
-	sfd = open(STOREF, O_RDWR | O_CREAT, 0600);
+	sfd = open(STOREF, flags, 0600);
 	if (sfd < 0)
 		err(1, "open %s", STOREF);
 
+	/*
+	 * The cache file does not have to exist
+	 * and will be created again if deleted.
+	 */
 	cfd = open(CACHEF, O_RDWR | O_CREAT, 0600);
 	if (cfd < 0)
 		err(1, "open %s", CACHEF);
@@ -475,58 +587,45 @@ init(void)
 	if (flock(ifd, LOCK_NB | LOCK_EX) < 0 ||
 	    flock(sfd, LOCK_NB | LOCK_EX) < 0 ||
 	    flock(cfd, LOCK_NB | LOCK_EX) < 0)
-		errx(1, "busy lock");
+		err(1, "flock");
 
-	if (fstat(ifd, &sb) < 0)
-		err(1, "fstat %s", SNAPSF);
-	if (sb.st_size != 0) {
-		uint8_t maj, min;
-
-		xread(ifd, &snaphdr, sizeof(snaphdr));
-		min = snaphdr.flags & 0xff;
-		maj = (snaphdr.flags >> 8) & 0xff;
-
-		if (maj != VER_MAJ || min != VER_MIN)
-			errx(1, "expected snapshot format version %u.%u but got %u.%u",
-			     VER_MAJ, VER_MIN, maj, min);
+	if (iflag) {
+		init_snap_hdr();
+		init_blk_hdr();
 	} else {
-		snaphdr.flags = (VER_MAJ << 8) | VER_MIN;
-		snaphdr.st.min_blk_size = comp_size(BLKSIZE_MAX);
-		xwrite(ifd, &snaphdr, sizeof(snaphdr));
+		load_snap_hdr();
+		load_blk_hdr();
 	}
 
 	cache = alloc_cache();
-	if (cache_nr_entries() != 0)
-		load_cache();
-	else
-		walk_snap(rebuild_cache, NULL);
+	load_cache();
 }
 
 static void
 term(void)
 {
-	if (verbose)
-		print_stats(&snaphdr.st);
+	if (verbose > 0)
+		print_stats(&snap_hdr.st);
 
-	if (cache_dirty) {
-		xlseek(cfd, 0, SEEK_SET);
-		walk_cache(cache, flush_cache);
-	}
+	save_cache();
 	free_cache(cache);
 
+	save_blk_hdr();
+	save_snap_hdr();
+
+	fsync(cfd);
+	fsync(sfd);
 	fsync(ifd);
-	fsync(sfd);
-	fsync(cfd);
 
-	close(ifd);
-	close(sfd);
 	close(cfd);
+	close(sfd);
+	close(ifd);
 }
 
 static void
 usage(void)
 {
-	fprintf(stderr, "usage: %s [-clv] [-e id] [-r root] [-m message] [file]\n", argv0);
+	fprintf(stderr, "usage: %s [-Zcilv] [-e id] [-r root] [-m message] [file]\n", argv0);
 	exit(1);
 }
 
@@ -535,15 +634,22 @@ main(int argc, char *argv[])
 {
 	uint8_t md[MDSIZE];
 	char *id = NULL, *root = NULL, *msg = NULL;
-	int fd = -1, lflag = 0, cflag = 0;
+	int iflag = 0, lflag = 0, cflag = 0;
+	int fd = -1;
 
 	ARGBEGIN {
+	case 'Z':
+		compr_enabled = 0;
+		break;
 	case 'c':
 		cflag = 1;
 		break;
 	case 'e':
 		id = EARGF(usage());
 		break;
+	case 'i':
+		iflag = 1;
+		break;
 	case 'l':
 		lflag = 1;
 		break;
@@ -554,7 +660,7 @@ main(int argc, char *argv[])
 		msg = EARGF(usage());
 		break;
 	case 'v':
-		verbose = 1;
+		verbose++;
 		break;
 	default:
 		usage();
@@ -585,15 +691,22 @@ main(int argc, char *argv[])
 			err(1, "chdir: %s", root);
 	}
 
-	init();
+	init(iflag);
+
+	if (iflag) {
+		term();
+		return 0;
+	}
 
 	if (cflag) {
+		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 		walk_snap(check, NULL);
 		term();
 		return 0;
 	}
 
 	if (lflag) {
+		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 		walk_snap(list, NULL);
 		term();
 		return 0;
@@ -602,6 +715,7 @@ main(int argc, char *argv[])
 	if (id) {
 		struct extract_args args;
 
+		xlseek(ifd, SNAP_HDR_SIZE, SEEK_SET);
 		str2bin(id, md);
 		args.md = md;
 		args.fd = fd;
